#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["NeptuneBackendMock"]

import uuid
from collections import defaultdict
from datetime import datetime
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from neptune.api.models import (
    BoolField,
    DateTimeField,
    Field,
    FieldDefinition,
    FieldType,
    FloatField,
    FloatPointValue,
    FloatSeriesField,
    FloatSeriesValues,
    IntField,
    LeaderboardEntry,
    NextPage,
    QueryFieldDefinitionsResult,
    QueryFieldsResult,
    StringField,
    StringPointValue,
    StringSeriesField,
    StringSeriesValues,
    StringSetField,
)
from neptune.core.components.operation_storage import OperationStorage
from neptune.exceptions import (
    ContainerUUIDNotFound,
    MetadataInconsistency,
    ModelVersionNotFound,
    ProjectNotFound,
    RunNotFound,
)
from neptune.internal.backends.api_model import (
    ApiExperiment,
    Project,
    Workspace,
)
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_structure import ContainerStructure
from neptune.internal.container_type import ContainerType
from neptune.internal.exceptions import (
    InternalClientError,
    NeptuneException,
)
from neptune.internal.id_formats import (
    QualifiedName,
    SysId,
    UniqueId,
)
from neptune.internal.operation import (
    AddStrings,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearFloatLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    CopyAttribute,
    DeleteAttribute,
    LogFloats,
    LogStrings,
    Operation,
    RemoveStrings,
)
from neptune.internal.operation_visitor import OperationVisitor
from neptune.internal.utils.generic_attribute_mapper import NoValue
from neptune.internal.utils.paths import path_to_str
from neptune.types import (
    Boolean,
    Integer,
)
from neptune.types.atoms.datetime import Datetime
from neptune.types.atoms.float import Float
from neptune.types.atoms.string import String
from neptune.types.namespace import Namespace
from neptune.types.series.float_series import FloatSeries
from neptune.types.series.string_series import StringSeries
from neptune.types.sets.string_set import StringSet
from neptune.types.value import Value
from neptune.types.value_visitor import ValueVisitor
from neptune.typing import ProgressBarType

Val = TypeVar("Val", bound=Value)


class NeptuneBackendMock(NeptuneBackend):
    WORKSPACE_NAME = "mock-workspace"
    PROJECT_NAME = "project-placeholder"
    PROJECT_KEY = SysId("OFFLINE")
    MODEL_SYS_ID = SysId("OFFLINE-MOD")

    def __init__(self, credentials=None, proxies=None):
        self._project_id: UniqueId = UniqueId(str(uuid.uuid4()))
        self._containers: Dict[(UniqueId, ContainerType), ContainerStructure[Value, dict]] = dict()
        self._next_run = 1  # counter for runs
        self._next_model_version = defaultdict(lambda: 1)  # counter for model versions
        self._attribute_type_converter_value_visitor = self.AttributeTypeConverterValueVisitor()
        self._create_container(self._project_id, ContainerType.PROJECT, self.PROJECT_KEY)

    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        return [
            Project(
                id=UniqueId(str(uuid.uuid4())),
                name=self.PROJECT_NAME,
                workspace=self.WORKSPACE_NAME,
                sys_id=self.PROJECT_KEY,
            )
        ]

    def get_available_workspaces(self) -> List[Workspace]:
        return [Workspace(id=UniqueId(str(uuid.uuid4())), name=self.WORKSPACE_NAME)]

    def _create_container(self, container_id: UniqueId, container_type: ContainerType, sys_id: SysId):
        container = self._containers.setdefault((container_id, container_type), ContainerStructure[Value, dict]())
        container.set(["sys", "id"], String(str(sys_id)))
        container.set(["sys", "state"], String("Active"))
        container.set(["sys", "owner"], String("offline_user"))
        container.set(["sys", "size"], Float(0))
        container.set(["sys", "tags"], StringSet(set()))
        container.set(["sys", "creation_time"], Datetime(datetime.now()))
        container.set(["sys", "modification_time"], Datetime(datetime.now()))
        container.set(["sys", "failed"], Boolean(False))
        if container_type == ContainerType.MODEL_VERSION:
            container.set(["sys", "model_id"], String(str(self.MODEL_SYS_ID)))
            container.set(["sys", "stage"], String("none"))
        return container

    def _get_container(self, container_id: UniqueId, container_type: ContainerType):
        key = (container_id, container_type)
        if key not in self._containers:
            raise ContainerUUIDNotFound(container_id, container_type)
        container = self._containers[(container_id, container_type)]
        return container

    def create_run(
        self,
        project_id: UniqueId,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiExperiment:
        sys_id = SysId(f"{self.PROJECT_KEY}-{self._next_run}")
        self._next_run += 1
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.RUN, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.RUN,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def create_model(self, project_id: str, key: str) -> ApiExperiment:
        sys_id = SysId(f"{self.PROJECT_KEY}-{key}")
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.MODEL, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.MODEL,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def create_model_version(self, project_id: str, model_id: UniqueId) -> ApiExperiment:
        try:
            model_key = self._get_container(container_id=model_id, container_type=ContainerType.MODEL).get("sys/id")
        except ContainerUUIDNotFound:
            model_key = "MOD"

        sys_id = SysId(f"{self.PROJECT_KEY}-{model_key}-{self._next_model_version[model_id]}")
        self._next_model_version[model_id] += 1
        new_run_id = UniqueId(str(uuid.uuid4()))
        self._create_container(new_run_id, ContainerType.MODEL_VERSION, sys_id=sys_id)
        return ApiExperiment(
            id=new_run_id,
            type=ContainerType.MODEL,
            sys_id=sys_id,
            workspace=self.WORKSPACE_NAME,
            project_name=self.PROJECT_NAME,
            trashed=False,
        )

    def get_project(self, project_id: QualifiedName) -> Project:
        return Project(
            id=self._project_id,
            name=self.PROJECT_NAME,
            workspace=self.WORKSPACE_NAME,
            sys_id=self.PROJECT_KEY,
        )

    def get_metadata_container(
        self,
        container_id: Union[UniqueId, QualifiedName],
        expected_container_type: Optional[ContainerType],
    ) -> ApiExperiment:
        if "/" not in container_id:
            raise ValueError("Backend mock expect container_id as QualifiedName only")

        if expected_container_type == ContainerType.RUN:
            raise RunNotFound(container_id)
        elif expected_container_type == ContainerType.MODEL:
            return ApiExperiment(
                id=UniqueId(str(uuid.uuid4())),
                type=ContainerType.MODEL,
                sys_id=SysId(container_id.rsplit("/", 1)[-1]),
                workspace=self.WORKSPACE_NAME,
                project_name=self.PROJECT_NAME,
            )
        elif expected_container_type == ContainerType.MODEL_VERSION:
            raise ModelVersionNotFound(container_id)
        else:
            raise ProjectNotFound(container_id)

    def execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
        operation_storage: OperationStorage,
    ) -> Tuple[int, List[NeptuneException]]:
        result = []
        for op in operations:
            try:
                self._execute_operation(container_id, container_type, op, operation_storage)
            except NeptuneException as e:
                result.append(e)
        return len(operations), result

    def _execute_operation(
        self, container_id: UniqueId, container_type: ContainerType, op: Operation, operation_storage: OperationStorage
    ) -> None:
        run = self._get_container(container_id, container_type)
        val = run.get(op.path)
        if val is not None and not isinstance(val, Value):
            if isinstance(val, dict):
                raise MetadataInconsistency("{} is a namespace, not an attribute".format(op.path))
            else:
                raise InternalClientError("{} is a {}".format(op.path, type(val)))
        visitor = NeptuneBackendMock.NewValueOpVisitor(self, op.path, val, operation_storage)
        new_val = visitor.visit(op)
        if new_val is not None:
            run.set(op.path, new_val)
        else:
            run.pop(op.path)

    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[FieldDefinition]:
        run = self._get_container(container_id, container_type)
        return list(self._generate_attributes(None, run.get_structure()))

    def _generate_attributes(self, base_path: Optional[str], values: dict):
        for key, value_or_dict in values.items():
            new_path = base_path + "/" + key if base_path is not None else key
            if isinstance(value_or_dict, dict):
                yield from self._generate_attributes(new_path, value_or_dict)
            else:
                yield FieldDefinition(
                    new_path,
                    value_or_dict.accept(self._attribute_type_converter_value_visitor),
                )

    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatField:
        val = self._get_attribute(container_id, container_type, path, Float)
        return FloatField(path=path_to_str(path), value=val.value)

    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntField:
        val = self._get_attribute(container_id, container_type, path, Integer)
        return IntField(path=path_to_str(path), value=val.value)

    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolField:
        val = self._get_attribute(container_id, container_type, path, Boolean)
        return BoolField(path=path_to_str(path), value=val.value)

    def get_string_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> StringField:
        val = self._get_attribute(container_id, container_type, path, String)
        return StringField(path=path_to_str(path), value=val.value)

    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DateTimeField:
        val = self._get_attribute(container_id, container_type, path, Datetime)
        return DateTimeField(path=path_to_str(path), value=val.value)

    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesField:
        val = self._get_attribute(container_id, container_type, path, FloatSeries)
        return FloatSeriesField(path=path_to_str(path), last=val.values[-1] if val.values else None)

    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesField:
        val = self._get_attribute(container_id, container_type, path, StringSeries)
        return StringSeriesField(path=path_to_str(path), last=val.values[-1] if val.values else None)

    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetField:
        val = self._get_attribute(container_id, container_type, path, StringSet)
        return StringSetField(path=path_to_str(path), values=set(val.values))

    def _get_attribute(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        expected_type: Type[Val],
    ) -> Val:
        run = self._get_container(container_id, container_type)
        value: Optional[Value] = run.get(path)
        str_path = path_to_str(path)
        if value is None:
            raise MetadataInconsistency("Attribute {} not found".format(str_path))
        if isinstance(value, expected_type):
            return value
        raise MetadataInconsistency("Attribute {} is not {}".format(str_path, type.__name__))

    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
    ) -> StringSeriesValues:
        val = self._get_attribute(container_id, container_type, path, StringSeries)
        return StringSeriesValues(
            len(val.values),
            [StringPointValue(timestamp=datetime.now(), step=idx, value=v) for idx, v in enumerate(val.values)],
        )

    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
        use_proto: Optional[bool] = None,
    ) -> FloatSeriesValues:
        val = self._get_attribute(container_id, container_type, path, FloatSeries)
        return FloatSeriesValues(
            len(val.values),
            [FloatPointValue(timestamp=datetime.now(), step=idx, value=v) for idx, v in enumerate(val.values)],
        )

    def get_fields_definitions(
        self,
        container_id: str,
        container_type: ContainerType,
        use_proto: Optional[bool] = None,
    ) -> List[FieldDefinition]:
        return []

    def _get_attribute_values(self, value_dict, path_prefix: List[str]):
        assert isinstance(value_dict, dict)
        for k, value in value_dict.items():
            if isinstance(value, dict):
                yield from self._get_attribute_values(value, path_prefix + [k])
            else:
                attr_type = value.accept(self._attribute_type_converter_value_visitor).value
                attr_path = "/".join(path_prefix + [k])
                if hasattr(value, "value"):
                    yield attr_path, attr_type, value.value
                else:
                    return attr_path, attr_type, NoValue

    def fetch_atom_attribute_values(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> List[Tuple[str, FieldType, Any]]:
        run = self._get_container(container_id, container_type)
        values = self._get_attribute_values(run.get(path), path)
        namespace_prefix = path_to_str(path)
        if namespace_prefix:
            # don't want to catch "ns/attribute/other" while looking for "ns/attr"
            namespace_prefix += "/"
        return [
            (full_path, attr_type, attr_value)
            for (full_path, attr_type, attr_value) in values
            if full_path.startswith(namespace_prefix)
        ]

    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[Iterable[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[Iterable[str]] = None,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
        use_proto: Optional[bool] = None,
    ) -> Generator[LeaderboardEntry, None, None]:
        """Non relevant for mock"""

    class AttributeTypeConverterValueVisitor(ValueVisitor[FieldType]):
        def visit_float(self, _: Float) -> FieldType:
            return FieldType.FLOAT

        def visit_integer(self, _: Integer) -> FieldType:
            return FieldType.INT

        def visit_boolean(self, _: Boolean) -> FieldType:
            return FieldType.BOOL

        def visit_string(self, _: String) -> FieldType:
            return FieldType.STRING

        def visit_datetime(self, _: Datetime) -> FieldType:
            return FieldType.DATETIME

        def visit_float_series(self, _: FloatSeries) -> FieldType:
            return FieldType.FLOAT_SERIES

        def visit_string_series(self, _: StringSeries) -> FieldType:
            return FieldType.STRING_SERIES

        def visit_string_set(self, _: StringSet) -> FieldType:
            return FieldType.STRING_SET

        def visit_namespace(self, _: Namespace) -> FieldType:
            raise NotImplementedError

        def copy_value(self, source_type: Type[FieldDefinition], source_path: List[str]) -> FieldType:
            raise NotImplementedError

    class NewValueOpVisitor(OperationVisitor[Optional[Value]]):
        def __init__(
            self, backend, path: List[str], current_value: Optional[Value], operation_storage: OperationStorage
        ):
            self._backend = backend
            self._path = path
            self._current_value = current_value
            self._operation_storage = operation_storage

        def visit_assign_float(self, op: AssignFloat) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Float):
                raise self._create_type_error("assign", Float.__name__)
            return Float(op.value)

        def visit_assign_int(self, op: AssignInt) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Integer):
                raise self._create_type_error("assign", Integer.__name__)
            return Integer(op.value)

        def visit_assign_bool(self, op: AssignBool) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Boolean):
                raise self._create_type_error("assign", Boolean.__name__)
            return Boolean(op.value)

        def visit_assign_string(self, op: AssignString) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, String):
                raise self._create_type_error("assign", String.__name__)
            return String(op.value)

        def visit_assign_datetime(self, op: AssignDatetime) -> Optional[Value]:
            if self._current_value is not None and not isinstance(self._current_value, Datetime):
                raise self._create_type_error("assign", Datetime.__name__)
            return Datetime(op.value)

        def visit_log_floats(self, op: LogFloats) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return FloatSeries(raw_values)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(
                self._current_value.values + raw_values,
                min=self._current_value.min,
                max=self._current_value.max,
                unit=self._current_value.unit,
            )

        def visit_log_strings(self, op: LogStrings) -> Optional[Value]:
            raw_values = [x.value for x in op.values]
            if self._current_value is None:
                return StringSeries(raw_values)
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("log", StringSeries.__name__)
            return StringSeries(self._current_value.values + raw_values)

        def visit_clear_float_log(self, op: ClearFloatLog) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([])
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("clear", FloatSeries.__name__)
            return FloatSeries(
                [],
                min=self._current_value.min,
                max=self._current_value.max,
                unit=self._current_value.unit,
            )

        def visit_clear_string_log(self, op: ClearStringLog) -> Optional[Value]:
            if self._current_value is None:
                return StringSeries([])
            if not isinstance(self._current_value, StringSeries):
                raise self._create_type_error("clear", StringSeries.__name__)
            return StringSeries([])

        def visit_config_float_series(self, op: ConfigFloatSeries) -> Optional[Value]:
            if self._current_value is None:
                return FloatSeries([], min=op.min, max=op.max, unit=op.unit)
            if not isinstance(self._current_value, FloatSeries):
                raise self._create_type_error("log", FloatSeries.__name__)
            return FloatSeries(self._current_value.values, min=op.min, max=op.max, unit=op.unit)

        def visit_add_strings(self, op: AddStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(op.values)
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("add", StringSet.__name__)
            return StringSet(self._current_value.values.union(op.values))

        def visit_remove_strings(self, op: RemoveStrings) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("remove", StringSet.__name__)
            return StringSet(self._current_value.values.difference(op.values))

        def visit_clear_string_set(self, op: ClearStringSet) -> Optional[Value]:
            if self._current_value is None:
                return StringSet(set())
            if not isinstance(self._current_value, StringSet):
                raise self._create_type_error("clear", StringSet.__name__)
            return StringSet(set())

        def visit_delete_attribute(self, op: DeleteAttribute) -> Optional[Value]:
            if self._current_value is None:
                raise MetadataInconsistency(
                    "Cannot perform delete operation on {}. Attribute is undefined.".format(self._path)
                )
            return None

        def visit_copy_attribute(self, op: CopyAttribute) -> Optional[Value]:
            return op.resolve(self._backend).accept(self)

        def _create_type_error(self, op_name, expected):
            return MetadataInconsistency(
                "Cannot perform {} operation on {}. Expected {}, {} found.".format(
                    op_name, self._path, expected, type(self._current_value)
                )
            )

    def get_fields_with_paths_filter(
        self, container_id: str, container_type: ContainerType, paths: List[str], use_proto: Optional[bool] = None
    ) -> List[Field]:
        return []

    def query_fields_definitions_within_project(
        self,
        project_id: QualifiedName,
        field_name_regex: Optional[str] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldDefinitionsResult:
        return QueryFieldDefinitionsResult(
            entries=[],
            next_page=NextPage(next_page_token=None, limit=0),
        )

    def query_fields_within_project(
        self,
        project_id: QualifiedName,
        field_names_filter: Optional[List[str]] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldsResult:
        return QueryFieldsResult(
            entries=[],
            next_page=NextPage(next_page_token=None, limit=0),
        )



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("paginate_over",)

import abc
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
)

from typing_extensions import Protocol

from neptune.api.models import NextPage


@dataclass
class WithPagination(abc.ABC):
    next_page: Optional[NextPage]


T = TypeVar("T", bound=WithPagination)
Entry = TypeVar("Entry")


class Paginatable(Protocol):
    def __call__(self, *, next_page: Optional[NextPage] = None, **kwargs: Any) -> Any: ...


def paginate_over(
    getter: Paginatable,
    extract_entries: Callable[[T], Iterable[Entry]],
    **kwargs: Any,
) -> Iterator[Entry]:
    """
    Generic approach to pagination via `NextPage`
    """
    data = getter(**kwargs, next_page=None)
    yield from extract_entries(data)

    while data.next_page is not None and data.next_page.next_page_token is not None:
        data = getter(**kwargs, next_page=data.next_page)
        yield from extract_entries(data)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ProcessorStopSignal",
    "ProcessorStopLogger",
    "ProcessorStopSignalData",
    "ProcessorStopSignalType",
]

import logging
from dataclasses import (
    dataclass,
    field,
)
from enum import Enum
from queue import Queue
from typing import Optional

CONNECTION_INTERRUPTED_MSG = (
    "We have been experiencing connection interruptions during your run."
    " Neptune client will now try to resume connection and sync data for the next"
    " %s seconds."
    " You can also kill this process and synchronize your data manually later"
    " using `neptune sync` command."
)

WAITING_FOR_OPERATIONS_MSG = (
    "Waiting for the remaining %s operations to synchronize with Neptune." " Do not kill this process."
)

SUCCESS_MSG = "All %s operations synced, thanks for waiting!"

SYNC_FAILURE_MSG = (
    "Failed to sync all operations in %s seconds."
    " You have %s operations saved on disk that can be manually synced"
    " using `neptune sync` command."
)

RECONNECT_FAILURE_MSG = (
    "Failed to reconnect with Neptune in %s seconds."
    " You have %s operations saved on disk that can be manually synced"
    " using `neptune sync` command."
)

STILL_WAITING_MSG = "Still waiting for the remaining %s operations" " (%.2f%% done). Please wait."


class ProcessorStopSignalType(Enum):
    CONNECTION_INTERRUPTED = "CONNECTION_INTERRUPTED"
    WAITING_FOR_OPERATIONS = "WAITING_FOR_OPERATIONS"
    SUCCESS = "SUCCESS"
    SYNC_FAILURE = "SYNC_FAILURE"
    RECONNECT_FAILURE = "RECONNECT_FAILURE"
    STILL_WAITING = "STILL_WAITING"


@dataclass
class ProcessorStopSignalData:
    processor_id: int = 0
    size_remaining: int = 0
    already_synced: int = 0
    already_synced_proc: float = 0.0
    seconds: float = 0.0
    max_reconnect_wait_time: float = 0.0


@dataclass
class ProcessorStopSignal:
    signal_type: ProcessorStopSignalType
    data: ProcessorStopSignalData = field(default_factory=ProcessorStopSignalData)


class ProcessorStopLogger:
    def __init__(
        self,
        processor_id: int,
        signal_queue: Optional["Queue[ProcessorStopSignal]"],
        logger: logging.Logger,
        should_print_logs: bool = True,
    ) -> None:
        self._id = processor_id
        self._signal_queue = signal_queue
        self._logger = logger
        self._should_print_logs = should_print_logs

    def set_processor_stop_signal_queue(self, signal_queue: Optional["Queue[ProcessorStopSignal]"]) -> None:
        self._signal_queue = signal_queue

    def log_connection_interruption(self, max_reconnect_wait_time: float) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.CONNECTION_INTERRUPTED,
                    data=ProcessorStopSignalData(
                        processor_id=self._id, max_reconnect_wait_time=max_reconnect_wait_time
                    ),
                )
            )
        else:
            self._logger.warning(
                CONNECTION_INTERRUPTED_MSG,
                max_reconnect_wait_time,
            )

    def log_remaining_operations(self, size_remaining: int) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.WAITING_FOR_OPERATIONS,
                    data=ProcessorStopSignalData(processor_id=self._id, size_remaining=size_remaining),
                )
            )
        else:
            if size_remaining:
                self._logger.info(
                    WAITING_FOR_OPERATIONS_MSG,
                    size_remaining,
                )

    def log_success(self, ops_synced: int) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.SUCCESS,
                    data=ProcessorStopSignalData(processor_id=self._id, already_synced=ops_synced),
                )
            )
        else:
            if self._should_print_logs:
                self._logger.info(SUCCESS_MSG, ops_synced)

    def log_sync_failure(self, seconds: float, size_remaining: int) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.SYNC_FAILURE,
                    data=ProcessorStopSignalData(processor_id=self._id, seconds=seconds),
                )
            )
        else:
            if self._should_print_logs:
                self._logger.warning(
                    SYNC_FAILURE_MSG,
                    seconds,
                    size_remaining,
                )

    def log_reconnect_failure(self, max_reconnect_wait_time: float, size_remaining: int) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.RECONNECT_FAILURE,
                    data=ProcessorStopSignalData(
                        processor_id=self._id,
                        max_reconnect_wait_time=max_reconnect_wait_time,
                        size_remaining=size_remaining,
                    ),
                )
            )
        else:
            if self._should_print_logs:
                self._logger.warning(
                    RECONNECT_FAILURE_MSG,
                    max_reconnect_wait_time,
                    size_remaining,
                )

    def log_still_waiting(self, size_remaining: int, already_synced: int, already_synced_proc: float) -> None:
        if self._signal_queue is not None:
            self._signal_queue.put(
                ProcessorStopSignal(
                    signal_type=ProcessorStopSignalType.STILL_WAITING,
                    data=ProcessorStopSignalData(
                        processor_id=self._id,
                        size_remaining=size_remaining,
                        already_synced=already_synced,
                        already_synced_proc=already_synced_proc,
                    ),
                )
            )
        else:
            if self._should_print_logs:
                self._logger.info(
                    STILL_WAITING_MSG,
                    size_remaining,
                    already_synced_proc,
                )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["SyncRunner"]

from pathlib import Path
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Sequence,
)

from neptune.cli.collect import collect_containers
from neptune.cli.utils import (
    get_metadata_container,
    get_project,
)
from neptune.constants import OFFLINE_NAME_PREFIX
from neptune.exceptions import CannotSynchronizeOfflineRunsWithoutProject
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.utils.logger import get_logger

if TYPE_CHECKING:
    from neptune.cli.containers import (
        AsyncContainer,
        OfflineContainer,
    )
    from neptune.internal.backends.neptune_backend import NeptuneBackend


logger = get_logger(with_prefix=False)


class SyncRunner:
    @staticmethod
    def sync_all_offline(*, backend: "NeptuneBackend", base_path: Path, project_name: Optional[str] = None) -> None:
        containers = collect_containers(path=base_path, backend=backend)

        project = get_project(project_name_flag=QualifiedName(project_name) if project_name else None, backend=backend)
        if not project:
            raise CannotSynchronizeOfflineRunsWithoutProject

        for container in containers.offline_containers:
            container.sync(base_path=base_path, backend=backend, project=project)

    @staticmethod
    def sync_all(*, backend: "NeptuneBackend", base_path: Path, project_name: Optional[str] = None) -> None:
        containers = collect_containers(path=base_path, backend=backend)

        if containers.unsynced_containers:
            for async_container in containers.unsynced_containers:
                async_container.sync(base_path=base_path, backend=backend, project=None)

        if containers.offline_containers:
            project = get_project(
                project_name_flag=QualifiedName(project_name) if project_name else None, backend=backend
            )
            if not project:
                raise CannotSynchronizeOfflineRunsWithoutProject

            for offline_container in containers.offline_containers:
                offline_container.sync(base_path=base_path, backend=backend, project=project)

    @staticmethod
    def sync_selected(
        *, backend: "NeptuneBackend", base_path: Path, project_name: Optional[str] = None, object_names: Sequence[str]
    ) -> None:
        containers = collect_containers(path=base_path, backend=backend)
        async_selected = [QualifiedName(name) for name in object_names if not name.startswith(OFFLINE_NAME_PREFIX)]

        if async_selected:
            sync_selected_async(
                backend=backend,
                base_path=base_path,
                container_names=async_selected,
                containers=containers.async_containers,
            )

        offline_selected = [
            UniqueId(name[len(OFFLINE_NAME_PREFIX) :]) for name in object_names if name.startswith(OFFLINE_NAME_PREFIX)
        ]
        if offline_selected:
            sync_selected_offline(
                backend=backend,
                base_path=base_path,
                container_names=offline_selected,
                containers=containers.offline_containers,
                project_name=project_name,
            )


def sync_selected_async(
    *,
    backend: "NeptuneBackend",
    base_path: Path,
    container_names: List["QualifiedName"],
    containers: List["AsyncContainer"],
) -> None:
    async_containers_ids = set()
    for container_name in container_names:
        experiment = get_metadata_container(
            backend=backend,
            container_id=container_name,
        )
        if experiment:
            async_containers_ids.add(experiment.id)
        else:
            logger.error(f"Container {container_name} not found")

    selected_async_containers = [x for x in containers if x.container_id in async_containers_ids]

    for container in selected_async_containers:
        container.sync(base_path=base_path, backend=backend, project=None)


def sync_selected_offline(
    *,
    backend: "NeptuneBackend",
    base_path: Path,
    container_names: List["UniqueId"],
    containers: List["OfflineContainer"],
    project_name: Optional[str] = None,
) -> None:
    project = get_project(project_name_flag=QualifiedName(project_name) if project_name else None, backend=backend)
    if not project:
        raise CannotSynchronizeOfflineRunsWithoutProject

    selected_offline_containers: List["OfflineContainer"] = []
    for container_id in container_names:
        found_container = next((x for x in containers if x.container_id == container_id), None)

        if found_container:
            selected_offline_containers.append(found_container)
        else:
            logger.warning("Offline container %s not found on disk.", container_id)

    for container in selected_offline_containers:
        container.sync(base_path=base_path, backend=backend, project=project)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["NeptuneBackend"]

import abc
from typing import (
    Any,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

from neptune.api.models import (
    BoolField,
    DateTimeField,
    Field,
    FieldDefinition,
    FieldType,
    FloatField,
    FloatSeriesField,
    FloatSeriesValues,
    IntField,
    LeaderboardEntry,
    NextPage,
    QueryFieldDefinitionsResult,
    QueryFieldsResult,
    StringField,
    StringSeriesField,
    StringSeriesValues,
    StringSetField,
)
from neptune.core.components.operation_storage import OperationStorage
from neptune.internal.backends.api_model import (
    ApiExperiment,
    Project,
    Workspace,
)
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_type import ContainerType
from neptune.internal.exceptions import NeptuneException
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.internal.operation import Operation
from neptune.internal.websockets.websockets_factory import WebsocketsFactory
from neptune.typing import ProgressBarType


class NeptuneBackend:
    def close(self) -> None:
        """No need for closing implementation"""

    def verify_feature_available(self, _: str) -> None:
        """
        this method makes sense only for backends interacting with server;
        it makes sure that a feature is supported in the backend version client interacts with
        """

    def websockets_factory(self, project_id: str, run_id: str) -> Optional[WebsocketsFactory]:
        return None

    @abc.abstractmethod
    def get_project(self, project_id: QualifiedName) -> Project:
        pass

    @abc.abstractmethod
    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        pass

    @abc.abstractmethod
    def get_available_workspaces(self) -> List[Workspace]:
        pass

    @abc.abstractmethod
    def create_run(
        self,
        project_id: UniqueId,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def create_model(
        self,
        project_id: UniqueId,
        key: str,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def create_model_version(
        self,
        project_id: UniqueId,
        model_id: UniqueId,
    ) -> ApiExperiment:
        pass

    @abc.abstractmethod
    def get_metadata_container(
        self,
        container_id: Union[UniqueId, QualifiedName],
        expected_container_type: Optional[ContainerType],
    ) -> ApiExperiment:
        pass

    def ping(self, container_id: str, container_type: ContainerType):
        """Do nothing by default"""

    @abc.abstractmethod
    def execute_operations(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        operations: List[Operation],
        operation_storage: OperationStorage,
    ) -> Tuple[int, List[NeptuneException]]:
        pass

    @abc.abstractmethod
    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[FieldDefinition]:
        pass

    @abc.abstractmethod
    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatField:
        pass

    @abc.abstractmethod
    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntField:
        pass

    @abc.abstractmethod
    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolField:
        pass

    @abc.abstractmethod
    def get_string_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> StringField:
        pass

    @abc.abstractmethod
    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DateTimeField:
        pass

    @abc.abstractmethod
    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesField:
        pass

    @abc.abstractmethod
    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesField:
        pass

    @abc.abstractmethod
    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetField:
        pass

    @abc.abstractmethod
    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
    ) -> StringSeriesValues: ...

    @abc.abstractmethod
    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
        use_proto: Optional[bool] = None,
    ) -> FloatSeriesValues: ...

    # WARN: Used in Neptune Fetcher
    @abc.abstractmethod
    def get_fields_definitions(
        self,
        container_id: str,
        container_type: ContainerType,
        use_proto: Optional[bool] = None,
    ) -> List[FieldDefinition]: ...

    # WARN: Used in Neptune Fetcher
    @abc.abstractmethod
    def get_fields_with_paths_filter(
        self, container_id: str, container_type: ContainerType, paths: List[str], use_proto: Optional[bool] = None
    ) -> List[Field]: ...

    @abc.abstractmethod
    def fetch_atom_attribute_values(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> List[Tuple[str, FieldType, Any]]:
        pass

    @abc.abstractmethod
    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[List[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[List[str]] = None,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
        use_proto: Optional[bool] = None,
    ) -> Generator[LeaderboardEntry, None, None]:
        pass

    @abc.abstractmethod
    def query_fields_definitions_within_project(
        self,
        project_id: QualifiedName,
        field_name_regex: Optional[str] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldDefinitionsResult: ...

    @abc.abstractmethod
    def query_fields_within_project(
        self,
        project_id: QualifiedName,
        field_names_filter: Optional[List[str]] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldsResult: ...



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import neptune.api.operations as api_operations
from neptune.api.operations import Serializable
from neptune.core.operations import (
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    LogFloats,
)
from neptune.core.operations.operation import RunCreation
from neptune.core.operations.operation_visitor import OperationVisitor
from neptune.internal.utils.paths import path_to_str


class OperationToApiVisitor(OperationVisitor[Serializable]):
    def visit_assign_float(self, op: AssignFloat) -> Serializable:
        return api_operations.AssignFloat(path_to_str(op.path), op.value)

    def visit_assign_int(self, op: AssignInt) -> Serializable:
        return api_operations.AssignInteger(path_to_str(op.path), op.value)

    def visit_assign_bool(self, op: AssignBool) -> Serializable:
        return api_operations.AssignBool(path_to_str(op.path), op.value)

    def visit_assign_datetime(self, op: AssignDatetime) -> Serializable:
        return api_operations.AssignDatetime(path_to_str(op.path), op.value)

    def visit_assign_string(self, op: AssignString) -> Serializable:
        return api_operations.AssignString(path_to_str(op.path), op.value)

    def visit_log_floats(self, op: LogFloats) -> Serializable:
        return api_operations.LogFloats(
            path_to_str(op.path),
            [api_operations.FloatValue(val.ts, val.value, val.step) for val in op.values],
        )

    def visit_run_creation(self, op: RunCreation) -> Serializable:
        return api_operations.Run(op.created_at, op.custom_id)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("fetch_series_values",)

from typing import (
    Any,
    Callable,
    Iterator,
    Optional,
    TypeVar,
)

from neptune.api.models import (
    FloatPointValue,
    StringPointValue,
)
from neptune.internal.backends.utils import construct_progress_bar
from neptune.typing import ProgressBarType

PointValue = TypeVar("PointValue", StringPointValue, FloatPointValue)


def fetch_series_values(
    getter: Callable[..., Any], path: str, step_size: int = 1000, progress_bar: Optional[ProgressBarType] = None
) -> Iterator[PointValue]:
    first_batch = getter(from_step=None, limit=1)
    data_count = 0
    total = first_batch.total
    last_step_value = (first_batch.values[-1].step - 1) if first_batch.values else None
    progress_bar = False if total < step_size else progress_bar

    if total <= 1:
        yield from first_batch.values
        return

    with construct_progress_bar(progress_bar, f"Fetching {path} values") as bar:
        bar.update(by=data_count, total=total)

        while data_count < first_batch.total:
            batch = getter(from_step=last_step_value, limit=step_size)

            bar.update(by=len(batch.values), total=total)

            yield from batch.values

            last_step_value = batch.values[-1].step if batch.values else None
            data_count += len(batch.values)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

from neptune.core.components.operation_storage import OperationStorage
from neptune.exceptions import MalformedOperation
from neptune.internal.container_type import ContainerType

if TYPE_CHECKING:
    from neptune.attributes.attribute import Attribute
    from neptune.internal.backends.neptune_backend import NeptuneBackend
    from neptune.internal.operation_visitor import OperationVisitor

Ret = TypeVar("Ret")
T = TypeVar("T")


def all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])


@dataclass
class Operation(abc.ABC):

    path: List[str]

    @abc.abstractmethod
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        pass

    def clean(self, operation_storage: OperationStorage):
        pass

    def to_dict(self) -> dict:
        return {"type": self.__class__.__name__, "path": self.path}

    @staticmethod
    def from_dict(data: dict) -> "Operation":
        if "type" not in data:
            raise ValueError("Malformed operation {} - type is missing".format(data))
        sub_classes = {cls.__name__: cls for cls in all_subclasses(Operation)}
        if not data["type"] in sub_classes:
            raise ValueError("Malformed operation {} - unknown type {}".format(data, data["type"]))
        return sub_classes[data["type"]].from_dict(data)


@dataclass
class AssignFloat(Operation):

    value: float

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_float(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignFloat":
        return AssignFloat(data["path"], data["value"])


@dataclass
class AssignInt(Operation):

    value: int

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_int(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignInt":
        return AssignInt(data["path"], data["value"])


@dataclass
class AssignBool(Operation):

    value: bool

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_bool(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignBool":
        return AssignBool(data["path"], data["value"])


@dataclass
class AssignString(Operation):

    value: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_string(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignString":
        return AssignString(data["path"], data["value"])


@dataclass
class AssignDatetime(Operation):

    value: datetime

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_datetime(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = int(1000 * self.value.timestamp())
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AssignDatetime":
        return AssignDatetime(data["path"], datetime.fromtimestamp(data["value"] / 1000))


class LogOperation(Operation, abc.ABC):
    pass


@dataclass
class LogSeriesValue(Generic[T]):

    value: T
    step: Optional[float]
    ts: float

    def to_dict(self, value_serializer=lambda x: x) -> dict:
        return {"value": value_serializer(self.value), "step": self.step, "ts": self.ts}

    @staticmethod
    def from_dict(data: dict, value_deserializer=lambda x: x) -> "LogSeriesValue[T]":
        return LogSeriesValue[T](value_deserializer(data["value"]), data.get("step", None), data["ts"])


@dataclass
class LogFloats(LogOperation):

    ValueType = LogSeriesValue[float]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_floats(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> "LogFloats":
        return LogFloats(
            data["path"],
            [LogFloats.ValueType.from_dict(value) for value in data["values"]],
        )


@dataclass
class LogStrings(LogOperation):

    ValueType = LogSeriesValue[str]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> "LogStrings":
        return LogStrings(
            data["path"],
            [LogStrings.ValueType.from_dict(value) for value in data["values"]],
        )


@dataclass
class ClearFloatLog(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_float_log(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearFloatLog":
        return ClearFloatLog(data["path"])


@dataclass
class ClearStringLog(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_string_log(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearStringLog":
        return ClearStringLog(data["path"])


@dataclass
class ConfigFloatSeries(Operation):

    min: Optional[float]
    max: Optional[float]
    unit: Optional[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_config_float_series(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["min"] = self.min
        ret["max"] = self.max
        ret["unit"] = self.unit
        return ret

    @staticmethod
    def from_dict(data: dict) -> "ConfigFloatSeries":
        return ConfigFloatSeries(data["path"], data["min"], data["max"], data["unit"])


@dataclass
class AddStrings(Operation):

    values: Set[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_add_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "AddStrings":
        return AddStrings(data["path"], set(data["values"]))


@dataclass
class RemoveStrings(Operation):

    values: Set[str]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_remove_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> "RemoveStrings":
        return RemoveStrings(data["path"], set(data["values"]))


@dataclass
class ClearStringSet(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_clear_string_set(self)

    @staticmethod
    def from_dict(data: dict) -> "ClearStringSet":
        return ClearStringSet(data["path"])


@dataclass
class DeleteAttribute(Operation):
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_delete_attribute(self)

    @staticmethod
    def from_dict(data: dict) -> "DeleteAttribute":
        return DeleteAttribute(data["path"])


@dataclass
class CopyAttribute(Operation):
    container_id: str
    container_type: ContainerType
    source_path: List[str]
    source_attr_cls: Type["Attribute"]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_copy_attribute(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["container_id"] = self.container_id
        ret["container_type"] = self.container_type.value
        ret["source_path"] = self.source_path
        ret["source_attr_name"] = self.source_attr_cls.__name__
        return ret

    @staticmethod
    def from_dict(data: dict) -> "CopyAttribute":
        from neptune.attributes.attribute import Attribute

        source_attr_cls = {cls.__name__: cls for cls in all_subclasses(Attribute) if cls.supports_copy}.get(
            data["source_attr_name"]
        )

        if source_attr_cls is None:
            raise MalformedOperation("Copy of non-copiable type found in queue!")

        return CopyAttribute(
            data["path"],
            data["container_id"],
            ContainerType(data["container_type"]),
            data["source_path"],
            source_attr_cls,
        )

    def resolve(self, backend: "NeptuneBackend") -> Operation:
        # repack CopyAttribute op into target attribute assignment
        getter = self.source_attr_cls.getter
        create_assignment_operation = self.source_attr_cls.create_assignment_operation
        value = getter(backend, self.container_id, self.container_type, self.source_path)
        return create_assignment_operation(self.path, value)



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ("AsyncOperationProcessor",)

import threading
from pathlib import Path
from queue import Queue
from time import time
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
)

from neptune.core.components.abstract import (
    Resource,
    WithResources,
)
from neptune.core.operation_processors.async_operation_processor.constants import (
    STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS,
)
from neptune.core.operation_processors.async_operation_processor.consumer_thread import ConsumerThread
from neptune.core.operation_processors.async_operation_processor.operation_logger import ProcessorStopSignal
from neptune.core.operation_processors.async_operation_processor.processing_resources import ProcessingResources
from neptune.core.operation_processors.async_operation_processor.queue_observer import QueueObserver
from neptune.core.operation_processors.operation_processor import OperationProcessor
from neptune.core.operations.operation import Operation
from neptune.core.operations.utils import try_get_step
from neptune.core.typing.container_type import ContainerType
from neptune.core.typing.id_formats import CustomId
from neptune.exceptions import NeptuneSynchronizationAlreadyStoppedException
from neptune.internal.signals_processing.signals import Signal
from neptune.internal.utils.disk_utilization import ensure_disk_not_overutilize
from neptune.internal.warnings import (
    NeptuneWarning,
    warn_once,
)


class AsyncOperationProcessor(WithResources, OperationProcessor):
    def __init__(
        self,
        custom_id: CustomId,
        container_type: ContainerType,
        lock: threading.RLock,
        signal_queue: "Queue[Signal]",
        batch_size: int = 1,
        data_path: Optional[Path] = None,
        serializer: Callable[[Operation], Dict[str, Any]] = lambda op: op.to_dict(),
        should_print_logs: bool = True,
        sleep_time: float = 5.0,
    ) -> None:
        self._should_print_logs = should_print_logs
        self._accepts_operations: bool = True
        self._last_version: int = 0

        self._processing_resources = ProcessingResources(
            batch_size=batch_size,
            custom_id=custom_id,
            container_type=container_type,
            lock=lock,
            signal_queue=signal_queue,
            data_path=data_path,
            serializer=serializer,
        )

        self._consumer = ConsumerThread(
            sleep_time=sleep_time,
            processing_resources=self._processing_resources,
        )

        self._queue_observer = QueueObserver(
            disk_queue=self._processing_resources.disk_queue,
            consumer=self._consumer,
            should_print_logs=self._should_print_logs,
            stop_queue_max_time_no_connection_seconds=STOP_QUEUE_MAX_TIME_NO_CONNECTION_SECONDS,
        )

    @property
    def resources(self) -> Tuple[Resource, ...]:
        return self._processing_resources.resources

    @property
    def data_path(self) -> Path:
        return self._processing_resources.data_path

    @property
    def processing_resources(self) -> "ProcessingResources":
        return self._processing_resources

    @ensure_disk_not_overutilize
    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        if not self._accepts_operations:
            warn_once("Not accepting operations", exception=NeptuneWarning)
            return

        step = try_get_step(op)
        self._last_version = self.processing_resources.disk_queue.put(op, category=step)

        if _queue_has_enough_space(self.processing_resources.disk_queue.size(), self._processing_resources.batch_size):
            self._consumer.wake_up()
        if wait:
            self.wait()

    def start(self) -> None:
        self._consumer.start()

    def pause(self) -> None:
        self._consumer.pause()
        self.flush()

    def resume(self) -> None:
        self._consumer.resume()

    def wait(self) -> None:
        self.flush()
        waiting_for_version = self._last_version
        self._consumer.wake_up()

        # Probably reentering lock just for sure
        with self._processing_resources.waiting_cond:
            self._processing_resources.waiting_cond.wait_for(
                lambda: self._processing_resources.consumed_version >= waiting_for_version
                or not self._consumer.is_running()
            )
        if not self._consumer.is_running():
            raise NeptuneSynchronizationAlreadyStoppedException()

    def stop(
        self,
        seconds: Optional[float] = None,
        processor_stop_signal_queue: Optional["Queue[ProcessorStopSignal]"] = None,
    ) -> None:
        ts = time()
        self.flush()
        if self._consumer.is_running():
            self._consumer.disable_sleep()
            self._consumer.wake_up()
            self._queue_observer.wait_for_queue_empty(
                seconds=seconds,
                processor_stop_signal_queue=processor_stop_signal_queue,
            )
            self._consumer.interrupt()
        sec_left = None if seconds is None else seconds - (time() - ts)
        self._consumer.join(sec_left)

        # Close resources
        self.close()

        # Remove local files
        if self._queue_observer.is_queue_empty():
            self.cleanup()

    def cleanup(self) -> None:
        self._processing_resources.cleanup()

    def close(self) -> None:
        self._accepts_operations = False
        super().close()


def _queue_has_enough_space(queue_size: int, batch_size: int) -> bool:
    return queue_size > batch_size / 2



#
# Copyright (c) 2024, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "AssignInt",
    "AssignFloat",
    "AssignBool",
    "AssignString",
    "AssignDatetime",
    "LogFloats",
    "Operation",
    "FieldOperation",
    "RunCreation",
]

import abc
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from neptune.core.components.operation_storage import OperationStorage
from neptune.exceptions import MalformedOperation

if TYPE_CHECKING:
    from neptune.core.operations.operation_visitor import OperationVisitor

Ret = TypeVar("Ret")
T = TypeVar("T")


@dataclass
class Operation(abc.ABC):
    _registry: ClassVar[Dict[str, Type["Operation"]]] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        cls._registry[cls.__name__] = cls

    @abc.abstractmethod
    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        pass

    def clean(self, operation_storage: OperationStorage) -> None:
        pass

    def to_dict(self) -> Dict[str, Any]:
        return {"type": self.__class__.__name__}

    @classmethod
    def from_dict(cls, data: dict) -> "Operation":
        if "type" not in data:
            raise MalformedOperation("Malformed operation {} - type is missing".format(data))
        operation_type = data["type"]
        if operation_type not in Operation._registry:
            raise MalformedOperation("Malformed operation {} - unknown type {}".format(data, operation_type))
        return Operation._registry[operation_type].from_dict(data)


@dataclass
class RunCreation(Operation):
    created_at: datetime
    custom_id: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_run_creation(self)

    @classmethod
    def from_dict(cls, data: dict) -> "Operation":
        return cls(data["created_at"], data["custom_id"])

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), "created_at": self.created_at, "custom_id": self.custom_id}


@dataclass
class FieldOperation(Operation, abc.ABC):
    path: List[str]

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), "path": self.path}


@dataclass
class AssignFloat(FieldOperation):
    value: float

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_float(self)

    def to_dict(self) -> dict:
        return {**super().to_dict(), "value": self.value}

    @classmethod
    def from_dict(cls, data: dict) -> "AssignFloat":
        return cls(data["path"], data["value"])


@dataclass
class AssignInt(FieldOperation):
    value: int

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_int(self)

    def to_dict(self) -> dict:
        return {**super().to_dict(), "value": self.value}

    @classmethod
    def from_dict(cls, data: dict) -> "AssignInt":
        return cls(data["path"], data["value"])


@dataclass
class AssignBool(FieldOperation):
    value: bool

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_bool(self)

    def to_dict(self) -> dict:
        return {**super().to_dict(), "value": self.value}

    @classmethod
    def from_dict(cls, data: dict) -> "AssignBool":
        return cls(data["path"], data["value"])


@dataclass
class AssignString(FieldOperation):
    value: str

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_string(self)

    def to_dict(self) -> dict:
        return {**super().to_dict(), "value": self.value}

    @classmethod
    def from_dict(cls, data: dict) -> "AssignString":
        return AssignString(data["path"], data["value"])


@dataclass
class AssignDatetime(FieldOperation):
    value: datetime

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_assign_datetime(self)

    def to_dict(self) -> dict:
        return {**super().to_dict(), "value": int(1000 * self.value.timestamp())}

    @classmethod
    def from_dict(cls, data: dict) -> "AssignDatetime":
        return AssignDatetime(data["path"], datetime.fromtimestamp(data["value"] / 1000))


class LogOperation(FieldOperation, abc.ABC):
    pass


@dataclass
class LogSeriesValue(Generic[T]):
    value: T
    step: Optional[float]
    ts: float

    def to_dict(
        self,
        value_serializer: Callable[[T], Any] = lambda x: x,
    ) -> Dict[str, Union[T, Optional[float], float]]:
        return {"value": value_serializer(self.value), "step": self.step, "ts": self.ts}

    @staticmethod
    def from_dict(data: Dict[str, Any], value_deserializer: Callable[[T], Any] = lambda x: x) -> "LogSeriesValue[T]":
        return LogSeriesValue[T](value_deserializer(data["value"]), data.get("step"), data["ts"])


@dataclass
class LogFloats(LogOperation):
    ValueType = LogSeriesValue[float]

    values: List[ValueType]

    def accept(self, visitor: "OperationVisitor[Ret]") -> Ret:
        return visitor.visit_log_floats(self)

    def to_dict(self) -> Dict[str, Any]:
        return {**super().to_dict(), "values": [value.to_dict() for value in self.values]}

    @classmethod
    def from_dict(cls, data: dict) -> "LogFloats":
        return cls(
            data["path"],
            [cls.ValueType.from_dict(value) for value in data["values"]],  # type: ignore[misc]
        )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OfflineNeptuneBackend"]

from typing import (
    Generator,
    Iterable,
    List,
    Optional,
)

from neptune.api.models import (
    BoolField,
    DateTimeField,
    Field,
    FieldDefinition,
    FloatField,
    FloatSeriesField,
    FloatSeriesValues,
    IntField,
    LeaderboardEntry,
    NextPage,
    QueryFieldDefinitionsResult,
    QueryFieldsResult,
    StringField,
    StringSeriesField,
    StringSeriesValues,
    StringSetField,
)
from neptune.exceptions import NeptuneOfflineModeFetchException
from neptune.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.internal.backends.nql import NQLQuery
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import (
    QualifiedName,
    UniqueId,
)
from neptune.typing import ProgressBarType


class OfflineNeptuneBackend(NeptuneBackendMock):
    WORKSPACE_NAME = "offline"

    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[FieldDefinition]:
        raise NeptuneOfflineModeFetchException

    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatField:
        raise NeptuneOfflineModeFetchException

    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntField:
        raise NeptuneOfflineModeFetchException

    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolField:
        raise NeptuneOfflineModeFetchException

    def get_string_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> StringField:
        raise NeptuneOfflineModeFetchException

    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DateTimeField:
        raise NeptuneOfflineModeFetchException

    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesField:
        raise NeptuneOfflineModeFetchException

    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesField:
        raise NeptuneOfflineModeFetchException

    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetField:
        raise NeptuneOfflineModeFetchException

    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
    ) -> StringSeriesValues:
        raise NeptuneOfflineModeFetchException

    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        limit: int,
        from_step: Optional[float] = None,
        use_proto: Optional[bool] = None,
    ) -> FloatSeriesValues:
        raise NeptuneOfflineModeFetchException

    def get_fields_with_paths_filter(
        self, container_id: str, container_type: ContainerType, paths: List[str], use_proto: Optional[bool] = None
    ) -> List[Field]:
        raise NeptuneOfflineModeFetchException

    def get_fields_definitions(
        self,
        container_id: str,
        container_type: ContainerType,
        use_proto: Optional[bool] = None,
    ) -> List[FieldDefinition]:
        raise NeptuneOfflineModeFetchException

    def search_leaderboard_entries(
        self,
        project_id: UniqueId,
        types: Optional[Iterable[ContainerType]] = None,
        query: Optional[NQLQuery] = None,
        columns: Optional[Iterable[str]] = None,
        limit: Optional[int] = None,
        sort_by: str = "sys/creation_time",
        ascending: bool = False,
        progress_bar: Optional[ProgressBarType] = None,
        use_proto: Optional[bool] = None,
    ) -> Generator[LeaderboardEntry, None, None]:
        raise NeptuneOfflineModeFetchException

    def query_fields_definitions_within_project(
        self,
        project_id: QualifiedName,
        field_name_regex: Optional[str] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldDefinitionsResult:
        raise NeptuneOfflineModeFetchException

    def query_fields_within_project(
        self,
        project_id: QualifiedName,
        field_names_filter: Optional[List[str]] = None,
        experiment_ids_filter: Optional[List[str]] = None,
        next_page: Optional[NextPage] = None,
    ) -> QueryFieldsResult:
        raise NeptuneOfflineModeFetchException



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model/leaderboard_entries.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1fmodel/leaderboard_entries.proto\x12\x11neptune.api.model\"\xb3\x01\n&ProtoLeaderboardEntriesSearchResultDTO\x12\x1b\n\x13matching_item_count\x18\x01 \x01(\x03\x12\x1e\n\x11total_group_count\x18\x02 \x01(\x03H\x00\x88\x01\x01\x12\x36\n\x07\x65ntries\x18\x03 \x03(\x0b\x32%.neptune.api.model.ProtoAttributesDTOB\x14\n\x12_total_group_count\"\xd1\x01\n\x12ProtoAttributesDTO\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x12\n\nproject_id\x18\x03 \x01(\t\x12\x17\n\x0forganization_id\x18\x04 \x01(\t\x12\x14\n\x0cproject_name\x18\x05 \x01(\t\x12\x19\n\x11organization_name\x18\x06 \x01(\t\x12\x38\n\nattributes\x18\x07 \x03(\x0b\x32$.neptune.api.model.ProtoAttributeDTO\"\xed\x05\n\x11ProtoAttributeDTO\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x44\n\x0eint_properties\x18\x03 \x01(\x0b\x32\'.neptune.api.model.ProtoIntAttributeDTOH\x00\x88\x01\x01\x12H\n\x10\x66loat_properties\x18\x04 \x01(\x0b\x32).neptune.api.model.ProtoFloatAttributeDTOH\x01\x88\x01\x01\x12J\n\x11string_properties\x18\x05 \x01(\x0b\x32*.neptune.api.model.ProtoStringAttributeDTOH\x02\x88\x01\x01\x12\x46\n\x0f\x62ool_properties\x18\x06 \x01(\x0b\x32(.neptune.api.model.ProtoBoolAttributeDTOH\x03\x88\x01\x01\x12N\n\x13\x64\x61tetime_properties\x18\x07 \x01(\x0b\x32,.neptune.api.model.ProtoDatetimeAttributeDTOH\x04\x88\x01\x01\x12Q\n\x15string_set_properties\x18\x08 \x01(\x0b\x32-.neptune.api.model.ProtoStringSetAttributeDTOH\x05\x88\x01\x01\x12U\n\x17\x66loat_series_properties\x18\t \x01(\x0b\x32/.neptune.api.model.ProtoFloatSeriesAttributeDTOH\x06\x88\x01\x01\x42\x11\n\x0f_int_propertiesB\x13\n\x11_float_propertiesB\x14\n\x12_string_propertiesB\x12\n\x10_bool_propertiesB\x16\n\x14_datetime_propertiesB\x18\n\x16_string_set_propertiesB\x1a\n\x18_float_series_properties\"U\n\x14ProtoIntAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x03\"W\n\x16ProtoFloatAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x01\"X\n\x17ProtoStringAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"V\n\x15ProtoBoolAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x08\"Z\n\x19ProtoDatetimeAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x03\"[\n\x1aProtoStringSetAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x03(\t\"\xd1\x02\n\x1cProtoFloatSeriesAttributeDTO\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x16\n\x0e\x61ttribute_type\x18\x02 \x01(\t\x12\x16\n\tlast_step\x18\x03 \x01(\x01H\x00\x88\x01\x01\x12\x11\n\x04last\x18\x04 \x01(\x01H\x01\x88\x01\x01\x12\x10\n\x03min\x18\x05 \x01(\x01H\x02\x88\x01\x01\x12\x10\n\x03max\x18\x06 \x01(\x01H\x03\x88\x01\x01\x12\x14\n\x07\x61verage\x18\x07 \x01(\x01H\x04\x88\x01\x01\x12\x15\n\x08variance\x18\x08 \x01(\x01H\x05\x88\x01\x01\x12\x45\n\x06\x63onfig\x18\t \x01(\x0b\x32\x35.neptune.api.model.ProtoFloatSeriesAttributeConfigDTOB\x0c\n\n_last_stepB\x07\n\x05_lastB\x06\n\x04_minB\x06\n\x04_maxB\n\n\x08_averageB\x0b\n\t_variance\"t\n\"ProtoFloatSeriesAttributeConfigDTO\x12\x10\n\x03min\x18\x01 \x01(\x01H\x00\x88\x01\x01\x12\x10\n\x03max\x18\x02 \x01(\x01H\x01\x88\x01\x01\x12\x11\n\x04unit\x18\x03 \x01(\tH\x02\x88\x01\x01\x42\x06\n\x04_minB\x06\n\x04_maxB\x07\n\x05_unitB4\n0ml.neptune.leaderboard.api.model.proto.generatedP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'model.leaderboard_entries_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n0ml.neptune.leaderboard.api.model.proto.generatedP\001'
  _globals['_PROTOLEADERBOARDENTRIESSEARCHRESULTDTO']._serialized_start=55
  _globals['_PROTOLEADERBOARDENTRIESSEARCHRESULTDTO']._serialized_end=234
  _globals['_PROTOATTRIBUTESDTO']._serialized_start=237
  _globals['_PROTOATTRIBUTESDTO']._serialized_end=446
  _globals['_PROTOATTRIBUTEDTO']._serialized_start=449
  _globals['_PROTOATTRIBUTEDTO']._serialized_end=1198
  _globals['_PROTOINTATTRIBUTEDTO']._serialized_start=1200
  _globals['_PROTOINTATTRIBUTEDTO']._serialized_end=1285
  _globals['_PROTOFLOATATTRIBUTEDTO']._serialized_start=1287
  _globals['_PROTOFLOATATTRIBUTEDTO']._serialized_end=1374
  _globals['_PROTOSTRINGATTRIBUTEDTO']._serialized_start=1376
  _globals['_PROTOSTRINGATTRIBUTEDTO']._serialized_end=1464
  _globals['_PROTOBOOLATTRIBUTEDTO']._serialized_start=1466
  _globals['_PROTOBOOLATTRIBUTEDTO']._serialized_end=1552
  _globals['_PROTODATETIMEATTRIBUTEDTO']._serialized_start=1554
  _globals['_PROTODATETIMEATTRIBUTEDTO']._serialized_end=1644
  _globals['_PROTOSTRINGSETATTRIBUTEDTO']._serialized_start=1646
  _globals['_PROTOSTRINGSETATTRIBUTEDTO']._serialized_end=1737
  _globals['_PROTOFLOATSERIESATTRIBUTEDTO']._serialized_start=1740
  _globals['_PROTOFLOATSERIESATTRIBUTEDTO']._serialized_end=2077
  _globals['_PROTOFLOATSERIESATTRIBUTECONFIGDTO']._serialized_start=2079
  _globals['_PROTOFLOATSERIESATTRIBUTECONFIGDTO']._serialized_end=2195
# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["QueueElement", "DiskQueue"]

import json
import os
import threading
from collections import deque
from dataclasses import dataclass
from glob import glob
from pathlib import Path
from time import time
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Callable,
    Deque,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from neptune.core.components.abstract import WithResources
from neptune.core.components.queue.json_file_splitter import JsonFileSplitter
from neptune.core.components.queue.log_file import LogFile
from neptune.core.components.queue.sync_offset_file import SyncOffsetFile
from neptune.exceptions import MalformedOperation
from neptune.internal.utils.logger import get_logger

if TYPE_CHECKING:
    from neptune.core.components.abstract import Resource


T = TypeVar("T")
Timestamp = float

_logger = get_logger()


DEFAULT_MAX_BATCH_SIZE_BYTES = 100 * 1024**2


@dataclass
class QueueElement(Generic[T]):
    obj: T
    ver: int
    size: int
    at: Optional[Timestamp] = None


# NOTICE: This class is thread-safe as long as there is only one consumer and one producer.
class DiskQueue(WithResources, Generic[T]):
    def __init__(
        self,
        data_path: Path,
        to_dict: Callable[[T], dict],
        from_dict: Callable[[dict], T],
        lock: threading.RLock,
        max_file_size: int = 64 * 1024**2,
        max_batch_size_bytes: Optional[int] = None,
        extension: str = "log",
    ) -> None:
        self._data_path: Path = data_path.resolve()
        self._to_dict: Callable[[T], dict] = to_dict
        self._from_dict: Callable[[dict], T] = from_dict
        self._max_file_size: int = max_file_size
        self._max_batch_size_bytes: int = max_batch_size_bytes or int(
            os.environ.get("NEPTUNE_MAX_BATCH_SIZE_BYTES") or str(DEFAULT_MAX_BATCH_SIZE_BYTES)
        )
        self._extension: str = extension

        self._last_ack_file = SyncOffsetFile(data_path / "last_ack_version", default=0)
        self._last_put_file = SyncOffsetFile(data_path / "last_put_version", default=0)

        self._log_files: Deque[LogFile] = get_all_log_files(data_path, extension)
        self._write_file_version: int = self._log_files[-1].min_version
        self._writer = self._log_files[-1]
        self._read_file_version: int = self._log_files[0].min_version
        self._reader = JsonFileSplitter(self._log_files[0].file_path)

        self._should_skip_to_ack = True

        self._empty_cond = threading.Condition(lock)

    @property
    def data_path(self) -> Path:
        return self._data_path

    @property
    def resources(self) -> Tuple["Resource", ...]:
        return (
            self._last_put_file,
            self._last_ack_file,
        ) + tuple(self._log_files)

    def put(self, obj: T) -> int:
        version = self._last_put_file.read_local() + 1
        serialized_obj = json.dumps(self._serialize(obj=obj, version=version, at=time()))

        self._create_new_writer_if_file_size_exceeded(len(serialized_obj), version)

        self._writer.write(serialized_obj)
        self._last_put_file.write(version)

        return version

    def get(self) -> Optional[QueueElement[T]]:
        if self._should_skip_to_ack:
            return self._skip_and_get()
        else:
            return self._get()

    def _skip_and_get(self) -> Optional[QueueElement[T]]:
        ack_version = self._last_ack_file.read_local()
        while True:
            top_element = self._get()
            if top_element is None:
                return None
            if top_element.ver > ack_version:
                self._should_skip_to_ack = False
                if top_element.ver > ack_version + 1:
                    _logger.warning(
                        "Possible data loss. Last acknowledged operation version: %d, next: %d",
                        ack_version,
                        top_element.ver,
                    )
                return top_element

    def _get(self) -> Optional[QueueElement[T]]:
        _json, size = self._reader.get_with_size()
        if not _json:
            if self._read_file_version >= self._write_file_version:
                return None

            self._reader.close()
            for log_file in self._log_files:
                if log_file.min_version > self._read_file_version:
                    self._read_file_version = log_file.min_version
                    self._reader = JsonFileSplitter(log_file.file_path)
                    break

            # It is safe. Max recursion level is 2.
            return self._get()
        try:
            obj, ver, at = self._deserialize(_json)
            return QueueElement[T](obj, ver, size, at)
        except Exception as e:
            raise MalformedOperation from e

    def get_batch(self, size: int) -> List[QueueElement[T]]:
        if self._should_skip_to_ack:
            first = self._skip_and_get()
        else:
            first = self._get()
        if not first:
            return []

        ret = [first]
        cur_batch_size = first.size
        for _ in range(0, size - 1):
            if cur_batch_size >= self._max_batch_size_bytes:
                break
            next_obj = self._get()
            if not next_obj:
                break

            cur_batch_size += next_obj.size
            ret.append(next_obj)
        return ret

    def wait_for_empty(self, seconds: Optional[float] = None) -> bool:
        with self._empty_cond:
            return self._empty_cond.wait_for(self.is_empty, timeout=seconds)

    def ack(self, version: int) -> None:
        self._last_ack_file.write(version)
        self._clean_log_files_up_to(version)

        with self._empty_cond:
            if self.is_empty():
                self._empty_cond.notify_all()

    def _create_new_writer_if_file_size_exceeded(self, size: int, version: int) -> None:
        if self._writer.file_size + size > self._max_file_size:
            old_writer = self._writer
            self._writer = LogFile(self._data_path, version, extension=self._extension)
            old_writer.flush()
            old_writer.close()
            self._write_file_version = version
            self._log_files.append(self._writer)

    def _clean_log_files_up_to(self, version: int) -> None:
        log_versions = [log.min_version for log in self._log_files]

        for current_min_version, next_min_version in zip(log_versions, log_versions[1:]):
            if next_min_version <= version:
                self._log_files.popleft().cleanup()

    def is_empty(self) -> bool:
        return self.size() == 0

    def size(self) -> int:
        return self._last_put_file.read_local() - self._last_ack_file.read_local()

    def _serialize(self, obj: T, version: int, at: Optional[Timestamp] = None) -> dict:
        return {"obj": self._to_dict(obj), "version": version, "at": at}

    def _deserialize(self, data: dict) -> Tuple[T, int, Optional[Timestamp]]:
        return self._from_dict(data["obj"]), data["version"], data.get("at")

    def close(self) -> None:
        self._reader.close()
        super().close()

    def __enter__(self) -> "DiskQueue[T]":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.flush()
        self.close()
        if self.is_empty():
            self.cleanup()


def get_all_log_files(data_path: Path, extension: str) -> Deque[LogFile]:
    local_data_files = glob(f"{data_path}/data-*.{extension}")

    if not local_data_files:
        return deque([LogFile(data_path, 1, extension=extension)])

    sorted_local_data_files = sorted(
        local_data_files, key=lambda file_path: extract_version_from_file_name(Path(file_path), extension)
    )

    return deque(
        [
            LogFile(data_path, extract_version_from_file_name(Path(file_path), extension), extension=extension)
            for file_path in sorted_local_data_files
        ]
    )


def extract_version_from_file_name(file_path: Path, extension: str) -> int:
    return int(file_path.name.split("-")[-1][: -len(extension) - 1])



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["NoValue", "atomic_attribute_types_map", "map_attribute_result_to_value"]

from neptune.api.models import FieldType


class NoValue:
    pass


VALUE = "value"
LAST_VALUE = "last"
VALUES = "values"

atomic_attribute_types_map = {
    FieldType.FLOAT.value: "floatProperties",
    FieldType.INT.value: "intProperties",
    FieldType.BOOL.value: "boolProperties",
    FieldType.STRING.value: "stringProperties",
    FieldType.DATETIME.value: "datetimeProperties",
    FieldType.OBJECT_STATE.value: "experimentStateProperties",
    FieldType.NOTEBOOK_REF.value: "notebookRefProperties",
}

value_series_attribute_types_map = {
    FieldType.FLOAT_SERIES.value: "floatSeriesProperties",
    FieldType.STRING_SERIES.value: "stringSeriesProperties",
}

value_set_attribute_types_map = {
    FieldType.STRING_SET.value: "stringSetProperties",
}


def map_attribute_result_to_value(attribute):
    for attribute_map, value_key in [
        (atomic_attribute_types_map, VALUE),
        (value_series_attribute_types_map, LAST_VALUE),
        (value_set_attribute_types_map, VALUES),
    ]:
        source_property = attribute_map.get(attribute.type)
        if source_property is not None:
            mapped_attribute_entry = getattr(attribute, source_property)
            return getattr(mapped_attribute_entry, value_key)
    return NoValue



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ssl
import sys

from neptune.internal.utils.logger import get_logger

_logger = get_logger()

IS_WINDOWS = sys.platform == "win32"
IS_MACOS = sys.platform == "darwin"


def reset_internal_ssl_state():
    """
    OpenSSL's internal random number generator does not properly handle forked processes.
    Applications must change the PRNG state of the parent process if they use any SSL feature with os.fork().
    Any successful call of RAND_add(), RAND_bytes() or RAND_pseudo_bytes() is sufficient.
    https://docs.python.org/3/library/ssl.html#multi-processing
    """
    ssl.RAND_bytes(100)


def update_session_proxies(session, proxies):
    if proxies is not None:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def is_ipython():
    try:
        import IPython

        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False
